/*-
 * Copyright (c) 2003-2007 Tim Kientzle
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "test.h"
__FBSDID("$FreeBSD$");

#if defined(_WIN32) && !defined(__CYGWIN__)
#define open _open
#define lseek _lseek
#define close _close
#endif

DEFINE_TEST(test_open_fd)
{
    char buff[64];
    struct archive_entry *ae;
    struct archive *a;
    int fd;

    fd = open("test.tar", O_RDWR | O_CREAT | O_BINARY, 0777);
    assert(fd >= 0);
    if (fd < 0)
        return;

    /* Write an archive through this fd. */
    assert((a = archive_write_new()) != NULL);
    assertEqualIntA(a, ARCHIVE_OK, archive_write_set_format_ustar(a));
    assertEqualIntA(a, ARCHIVE_OK, archive_write_set_compression_none(a));
    assertEqualIntA(a, ARCHIVE_OK, archive_write_open_fd(a, fd));

    /*
     * Write a file to it.
     */
    assert((ae = archive_entry_new()) != NULL);
    archive_entry_set_mtime(ae, 1, 0);
    archive_entry_copy_pathname(ae, "file");
    archive_entry_set_mode(ae, S_IFREG | 0755);
    archive_entry_set_size(ae, 8);
    assertEqualIntA(a, ARCHIVE_OK, archive_write_header(a, ae));
    archive_entry_free(ae);
    assertEqualIntA(a, 8, archive_write_data(a, "12345678", 9));

    /*
     * Write a second file to it.
     */
    assert((ae = archive_entry_new()) != NULL);
    archive_entry_copy_pathname(ae, "file2");
    archive_entry_set_mode(ae, S_IFREG | 0755);
    archive_entry_set_size(ae, 819200);
    assertEqualIntA(a, ARCHIVE_OK, archive_write_header(a, ae));
    archive_entry_free(ae);

    /* Close out the archive. */
    assertEqualIntA(a, ARCHIVE_OK, archive_write_close(a));
    assertEqualInt(ARCHIVE_OK, archive_write_finish(a));

    /*
     * Now, read the data back.
     */
    assert(lseek(fd, 0, SEEK_SET) == 0);
    assert((a = archive_read_new()) != NULL);
    assertEqualIntA(a, ARCHIVE_OK, archive_read_support_format_all(a));
    assertEqualIntA(a, ARCHIVE_OK, archive_read_support_compression_all(a));
    assertEqualIntA(a, ARCHIVE_OK, archive_read_open_fd(a, fd, 512));

    assertEqualIntA(a, ARCHIVE_OK, archive_read_next_header(a, &ae));
    assertEqualInt(1, archive_entry_mtime(ae));
    assertEqualInt(0, archive_entry_mtime_nsec(ae));
    assertEqualInt(0, archive_entry_atime(ae));
    assertEqualInt(0, archive_entry_ctime(ae));
    assertEqualString("file", archive_entry_pathname(ae));
    assert((S_IFREG | 0755) == archive_entry_mode(ae));
    assertEqualInt(8, archive_entry_size(ae));
    assertEqualIntA(a, 8, archive_read_data(a, buff, 10));
    assertEqualMem(buff, "12345678", 8);

    assertEqualIntA(a, ARCHIVE_OK, archive_read_next_header(a, &ae));
    assertEqualString("file2", archive_entry_pathname(ae));
    assert((S_IFREG | 0755) == archive_entry_mode(ae));
    assertEqualInt(819200, archive_entry_size(ae));
    assertEqualIntA(a, ARCHIVE_OK, archive_read_data_skip(a));

    /* Verify the end of the archive. */
    assertEqualIntA(a, ARCHIVE_EOF, archive_read_next_header(a, &ae));
    assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
    assertEqualInt(ARCHIVE_OK, archive_read_finish(a));
    close(fd);


    /*
     * Verify some of the error handling.
     */
    assert((a = archive_read_new()) != NULL);
    assertEqualIntA(a, ARCHIVE_OK, archive_read_support_format_all(a));
    assertEqualIntA(a, ARCHIVE_OK, archive_read_support_compression_all(a));
    /* FD 100 shouldn't be open. */
    assertEqualIntA(a, ARCHIVE_FATAL,
        archive_read_open_fd(a, 100, 512));
    assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
    assertEqualInt(ARCHIVE_OK, archive_read_finish(a));
}
